/*
# soexp: written by Brane Zdralo brane.zdralo@fe.uni-lj.si <2005.07.04>
#    Program adjusts the pagesize of an EPS or single page PS file to fit content
# and converts it into a PNG or encapsulated PDF file. For this it requires
# Ghostscript6.0 or newer.
#    It also recognises Tango schematic and pcb PS plot files and makes possible
# to generate color PNG or PDF files out of them.
#    For Tango schematic Bill Of Materials report offers conversion into semicolon
# separated format suitable for inclusion into spreadsheet.
#    The way soexp makes conversion of EPS files is the same as in program
# epstopdf on which is based. The way soexp makes conversion of single page PS
# and Tango PS file is a little bit different.
#    Epstopdf was written by Kong Hoon Lee, konghoon@hyowon.cc.pusan.ac.kr <4/1/1999>
# and changed by Juergen Bausa, bausa@lfpt.rwth-aachen.de <1/11/2000>
# Epstopdf is based on the perl script 'epstopdf' by Sebastian Rahtz on
# http://tug.org/applications/pdftex/epstopdf .
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#define MAX_IN 10000

#ifndef GSEXEC
 #define GSEXEC "gs"
#endif


/* position of layers in input file */

typedef struct{
  long int layer_start;
  long int layer_end;
  int layer_gray;
}layerproto;


/* tango colors and graylevels */

typedef struct{
  double red;
  double green;
  double blue;
}colorproto;


/* default tango colors */

colorproto coltable[]={
    {0.00,0.00,0.00},     /* 1 black */
    {0.50,0.00,0.50},     /* 2 dark violet */
    {0.90,0.00,0.90},     /* 3 light violet */
    {1.00,1.00,1.00},     /* 4 white */
    {0.00,0.00,0.50},     /* 5 dark blue */
    {0.00,0.88,0.88},     /* 6 light cyan */
    {0.50,0.50,0.50},     /* 7 dark gray */
    {0.80,0.80,0.80},     /* 8 light gray */
    {1.00,0.50,0.00},     /* 9 orange */
    {1.00,1.00,0.00},     /* 10 yellow */
    {0.50,0.00,0.00},     /* 11 dark red */
    {0.90,0.00,0.00},     /* 12 light red */
    {0.00,0.50,0.00},     /* 13 dark green */
    {0.00,0.95,0.00},     /* 14 light green */
    {0.00,0.00,0.90},     /* 15 light blue*/
    {0.00,0.00,0.25},     /* 16 dark cyan */
  };


static void usage(void)

{
   printf("Usage: soexp [options] <input_file>\n");
   printf("Options:\n");
   printf(" -h, --help                 print usage\n");
   printf(" -i, --initfile=<file>      read settings from init file\n");
   printf(" -o, --outfile=<file>       write result to <file>\n");
   printf(" -f, --format=<pdf|png|eps> output format (default: pdf)\n");
   printf(" -d, --debug                debug informations (default: nodebug)\n");
   printf(" -g, --gsexec=<gs>          use <gs> to invoke ghostscript (default: %s)\n",GSEXEC);
   printf(" -t, --trim=<b>             shrink picture size to fits content and add\n");
   printf("                            border in inches (default: notrim)\n");
 }


int main(int argc,char *argv[])

{

  /* Postscript strings */

  const char psheaderstring[]="!PS-Adobe-";
  const char pagesstring[]="%%Pages:";
  const char pagestring[]="%%Page:";
  const char bbstring[]="%%BoundingBox:";
  const char creatorstring[]="%%Creator:";


   /* Windows PS file strings */

  const char mystring[]="/mysetup [";
  const char pagesizestring[]="dup /PageSize [";


  /* Tango strings */

  const char oldfontstring[]="Helvetica";
  const char graystring[]="setgray";
  const char colorstring[]="setrgbcolor";
  const char schstring[]="Schematic";
  const char pcbstring[]="PCB";
  const char mulstring[]="/m {";
  const char srstring[]="/sr {";
  const char prstring[]="/pr";


  /* Bill Of Material string  */

  const char bomstring[]="Bill of Materials";
  const char footerstring[]="Footer";
  const char totalstring[]="Total";

  FILE *in,*out;
  char *infile=NULL,*outfile=NULL,*initfile=NULL,*tempfile=NULL,*gsexec=NULL,*format=NULL;
  char *copt,*ptr,*bbfile;
  char command[300],ch[MAX_IN],ttemp[MAX_IN],newfontstring[300],pngcol[10];
  double dbbox[4],mul,temp[6],border,pngxdimm;
  double rout,gout,bout,thick,thin,newthick,newthin,lwidth;
  int bbox[4],width,height,xoffset,yoffset,page,pages,soffset,sstart;
  int i,count,status,lc,bom,pp,ss,ii,bc;
  char BBCorrected,did_CR,debug,compress,dopng,dopdf,trim,eps,ps;
  char pr,sr,tgocol,sch,pcb,pngxfactor;
  int layer,layer_N,layer_P,gray_temp;
  long int layer_temp,layer_temp_E,layer_1st;
  layerproto laytable[17];


  /* default values */

  status=0;
  debug=0;
  compress=1;
  dopdf=1;
  dopng=0;
  trim=0;
  bbfile="bbox.tmp";
  border=0;
  BBCorrected=0;
  pngxdimm=600;
  pngxfactor=0;

  bom=0;
  page=0;
  pages=1;
  eps=0;
  ps=0;
  lc=0;

  tgocol=1;
  layer=0;
  sch=0;
  pcb=0;
  thick=0;
  thin=0;
  newthick=1;
  newthin=1;
  lwidth=0;
  sr=0;
  pr=0;
  /* strcpy(newfontstring,"NimbusSanL-Regu");  old default font */
  strcpy(newfontstring,"URWGothicL-Demi");
  strcpy(pngcol,"png16m");


  /* Look up for init file and process it */

  for(i=1;i<argc;i++)
  {
    if(!strncmp(argv[i],"--initfile=",strlen("--initfile=")))
    {
      initfile=malloc(strlen(argv[i])-strlen("--initfile=")+1);
      strcpy(initfile,argv[i]+strlen("--initfile="));
    }
    else
    {
      if(!strcmp(argv[i],"-i") && i+1<argc)
      {
        initfile=malloc(strlen(argv[i+1])+1);
        strcpy(initfile, argv[i+1]);
      }
      else continue;
    }

    if(initfile != NULL)
    {
      if((in=fopen(initfile, "r")) != NULL)
      {
        while (fgets(ch,MAX_IN,in) != NULL )
        {
          if(ch[0]=='#' || ch[0] == '\n' || (ch[0] == '\r' && ch[1] == '\n')) continue;

          if(strncmp(ch,"pngcol",strlen("pngcol")) == 0)
          {
            sscanf(ch+strlen("pngcol"),"%s", pngcol);
            continue;
          }

          if(strncmp(ch,"gsexec",strlen("gsexec")) == 0)
          {
				gsexec=malloc(strlen(ch)-strlen("gsexec")-1);
				strncpy(gsexec,ch+strlen("gsexec")+1,strlen(ch)-strlen("gsexec")-2);
				if( gsexec[strlen(gsexec)-1] == '\r') gsexec[strlen(gsexec)-1]='\0'; 
				continue;
          }

          if(strncmp(ch,"format",strlen("format")) == 0)
          {
            sscanf(ch+strlen("format"),"%s", ttemp);
            if(!strcmp("png",ttemp))
            {
              dopdf=0;
              dopng=1;
              continue;
            }
            if(!strcmp("eps",ttemp))
            {
              dopdf=0;
              continue;
            }
            if(!strcmp("pdf",ttemp)) continue;
            printf("Unknown output format >>%s<<!\n", ttemp);
            printf("Using default output format pdf.\n");
            continue;
          }

          if(strncmp(ch,"trim",strlen("trim")) == 0)
          {
            sscanf(ch+strlen("trim"),"%lf", &border);
            trim=1;
            continue;
          }

          if(strncmp(ch,"nocompress",strlen("nocompress")) == 0)
          {
            compress=0;
            continue;
          }

          if(strncmp(ch,"debug",strlen("debug")) == 0)
          {
            debug=1;
            continue;
          }

          if(strncmp(ch,"notgocol",strlen("notgocol")) == 0)
          {
            tgocol=0;
            continue;
          }

          if(strncmp(ch,"color",strlen("color")) == 0)
          {
            sscanf(ch+strlen("color"),"%d %lf %lf %lf", &i,&rout,&gout,&bout);
            if(i < 15)
            {
              if(rout > 1) rout=1;
              if(gout > 1) gout=1;
              if(bout > 1) bout=1;
              coltable[i-1].red=rout;
              coltable[i-1].green=gout;
              coltable[i-1].blue=bout;
            }
            continue;
          }

          if(strncmp(ch,"font",strlen("font")) == 0)
          {
            sscanf(ch+strlen("font"),"%s", newfontstring);
            continue;
          }

          if(strncmp(ch,"thin",strlen("thin")) == 0)
          {
            sscanf(ch+strlen("thin"),"%lf", &newthin);
            continue;
          }

          if(strncmp(ch,"thick",strlen("thick")) == 0)
          {
            sscanf(ch+strlen("thick"),"%lf", &newthick);
            continue;
          }

          if(strncmp(ch,"pngxdimm",strlen("pngxdimm")) == 0)
          {
            sscanf(ch+strlen("pngxdimm"),"%lf", &pngxdimm);
            pngxfactor=0;
            continue;
          }

          if(strncmp(ch,"pngxfactor",strlen("pngxfactor")) == 0)
          {
            sscanf(ch+strlen("pngxfactor"),"%lf", &pngxdimm);
            pngxfactor=1;
            continue;
          }
          ch[strlen(ch)-2]='\0';
          printf("Warning: Unknown parameter >>%s<< !\n", ch);
        }
        fclose(in);
        free(initfile);
        break;
      }
      else printf("Warning: Initfile >>%s<< not found, continuing with defaults.\n", initfile);
      break;
    }
  }


  /* Process command line options */

  if(debug)printf("Processing command line...argc=%d\n",argc);

  for(i=1;i<argc;i++)
  {
    if(debug)printf("arg[%d]=>>%s<<\n",i,argv[i]);

    if(!strcmp(argv[i],"--help") || !strcmp(argv[i],"-h"))
    {
      usage();
      status=0;
      goto EXIT;
    }

    if(!strcmp(argv[i],"--debug") || !strcmp(argv[i],"-d"))
    {
      debug=1;
      continue;
    }

    if(!strncmp(argv[i],"--trim=",strlen("--trim=")))
    {
      trim=1;
      sscanf(argv[i]+strlen("--trim="),"%lf", &border);
      if(debug)printf("border %f\n", border);
      continue;
    }

    if(!strcmp(argv[i],"-t") && i+1<argc)
    {
      trim=1;
      i++;
      sscanf(argv[i],"%lf", &border);
      if(debug) printf("border %f\n", border);
      continue;
    }

    if(!strncmp(argv[i],"--outfile=",strlen("--outfile=")))
    {
      outfile=malloc(strlen(argv[i])-strlen("--outfile=")+1);
      strcpy(outfile, argv[i]+strlen("--outfile="));
      continue;
    }

    if(!strcmp(argv[i],"-o") && i+1<argc)
    {
      outfile=argv[i+1];
      i++;
      if(debug)printf("arg[%d]=>>%s<<\n",i,argv[i]);
      continue;
    }

    if(!strncmp(argv[i],"--initfile=",strlen("--initfile="))) continue;
    if(!strcmp(argv[i],"-i") && i+1<argc)
    {
      i++;
      if(debug)printf("arg[%d]=>>%s<<\n",i,argv[i]);
      continue;
    }

    if(!strncmp(argv[i],"--format=",strlen("--format=")))
    {
      format=malloc(strlen(argv[i])-strlen("--format=")+1);
      strcpy(format,argv[i]+strlen("--format="));
      continue;
    }

    if(!strcmp(argv[i],"-f") && i+1<argc)
    {
      format=(argv[i+1]);
      i++;
      if(debug)printf("arg[%d]=>>%s<<\n",i,argv[i]);
    }

    if(format != NULL)
    {
      if(!strcmp("pdf",format))
      {
        dopdf=1;
        dopng=0;
        format=NULL;
        continue;
      }

      if(!strcmp("png",format))
      {
        dopdf=0;
        dopng=1;
        format=NULL;
        continue;
      }

      if(!strcmp("eps",format))
      {
        dopdf=0;
        dopng=0;
        format=NULL;
        continue;
      }
      printf("ERROR: Unknown output file format! ");
      printf("Valid output formats are: pdf, png or eps.\n");
      status=1;
      goto EXIT;
    }

    if(!strncmp(argv[i],"--gsexec=",strlen("--gsexec=")))
    {
      if(gsexec != NULL) free(gsexec);
      gsexec=malloc(strlen(argv[i])-strlen("--gsexec=")+1);
      strcpy(gsexec, argv[i]+strlen("--gsexec="));
      continue;
    }

    if(!strcmp(argv[i],"-g") && i+1<argc)
    {
      gsexec=(argv[i+1]);
      i++;
      if(debug)printf("arg[%d]=>>%s<<\n",i,argv[i]);
      continue;
    }

    if(argv[i][0]!='-')
    {
      if(infile) printf("Warning: Only the first input file will be processed!\n");
      else       infile=argv[i];
      continue;
    }

    usage();
    printf("\nERROR: Unknown option >>%s<< !\n",argv[i]);
    goto EXIT;
  }


  if(!infile)
  {
    usage();
    printf("\nERROR: No input file specified!\n");
    status=1;
    goto EXIT;
  }


  if((in=fopen(infile,"r")) == NULL)
  {
    usage();
    printf("\nERROR: Input file >>%s<< not found!\n",infile);
    status=1;
    goto EXIT;
  }

  fclose(in);

  if(gsexec == NULL)
  {
    gsexec=malloc(strlen(GSEXEC)+1);
    strcpy(gsexec,GSEXEC);
  }


  /* find a temporary filename that does not exist yet */

  tempfile=malloc(strlen(infile)+8);
  count=0;
  do
  {
    if(count>99)
    {
      printf("No temporary file available!\n\n");
      status=1;
      goto EXIT;
    }
    sprintf(tempfile,"%s.%02d.tmp",infile,count);
    if(debug) printf("Checking temporary filename %s .\n",tempfile);
    out=fopen(tempfile,"r");
    if(out) fclose(out);
    count++;
  }while(out);


  /* test input file for format, creator, etc.. */

  in=fopen(infile, "r");

  while ((fgets(ch,MAX_IN,in)))
  {

    lc++;

    /* test input format - PS or EPS */

    if(strstr(ch,psheaderstring))
    {
      ps=1;
      if(strstr(ch,"EPSF"))
      {
        ps=0;
        eps=1;
      }
      continue;
    }


    /* test if Bill Of Material */

    if(strstr(ch,bomstring))
    {
      bom=1;
      break;
    }


    /* test if only one page - %%Pages 1 */

    if(!strncmp(ch,pagesstring,strlen(pagesstring)))
    {
      sscanf(ch+strlen(pagesstring),"%d",&pages);
      if(pages > 1)
      {
        fclose(in);
        printf("You can convert only single page documents!\n");
        status=1;
        goto EXIT;
      }
    }


    /* test if creator is Tango schematic or pcb */

    if (!strncmp(ch, creatorstring,strlen(creatorstring)))
    {
      if(strstr(ch,pcbstring)) pcb=1;
      if(strstr(ch,schstring)) sch=1;
      if((pcb || sch) && eps)
      {
        fclose(in);
        printf("In Tango you should use PS only!\n");
        status=1;
        goto EXIT;
      }
      continue;
    }


    /* For tango pcb record beginning of the layer(gaystring) and its graylevel */

    if(pcb)
    {
      if(page)
      {
        if(strstr(ch,graystring))
        {
          sscanf(ch+2,"%d", &i);
          laytable[layer].layer_gray=i/5;
          if(layer)laytable[layer-1].layer_end=layer_temp;
          else layer_1st=layer_temp;
          laytable[layer].layer_start=layer_temp;
          layer++;
        }
      }
      else
      {
        if(strncmp(ch,pagestring,strlen(pagestring)) == 0)
        {
          page=1;
          continue;
        }
      }
      layer_temp=ftell(in);
    }


    /* for Tango schematic find thick and thin line width */

    if(sch)
    {
      if(strstr(ch," pt"))
      {
        sscanf(strstr(ch," pt")-3,"%lf", &lwidth);
        if(lwidth != thick)
        {
          if(lwidth > thick)
          {
            thin=thick;
            thick=lwidth;
          }
          else thin=lwidth;
          continue;
        }
      }
    }
  }
  fclose(in);


  /* sort layers for tango pcb */

  if(pcb)
  {
    for(layer_N=0;layer_N < layer-1;layer_N++)
    {
      for(layer_P=0;layer_P < layer-1;layer_P++)
      {
        if(laytable[layer_P].layer_gray < laytable[layer_P+1].layer_gray)
        {
          layer_temp=laytable[layer_P].layer_start;
          layer_temp_E=laytable[layer_P].layer_end;
          gray_temp=laytable[layer_P].layer_gray;
          laytable[layer_P].layer_start=laytable[layer_P+1].layer_start;
          laytable[layer_P].layer_end=laytable[layer_P+1].layer_end;
          laytable[layer_P].layer_gray=laytable[layer_P+1].layer_gray;
          laytable[layer_P+1].layer_start=layer_temp;
          laytable[layer_P+1].layer_end=layer_temp_E;
          laytable[layer_P+1].layer_gray=gray_temp;
        }
      }
    }
  }


  /* generating output filename */

  if(!outfile)
  {
    outfile=malloc(strlen(infile)+7);
    strcpy(outfile,infile);
    ptr=strrchr(outfile,'.');
    if(ptr) *ptr='\0';
    if(bom) strcat(outfile,".kos");
    else if(dopdf) strcat(outfile,".pdf");
         else if(dopng) strcat(outfile,".png");
              else strcat(outfile,".2.eps");
  }


  /* processing Bill Of Material file */

  if(bom)
  {
    lc=0;
    bc=0;
    in=fopen(infile, "r");
    out=fopen(outfile,"w");

    while (fgets(ch,MAX_IN,in))
    {
      /* skip the first 9 lines */

      lc++;
      if(lc < 10) continue;

      if(strstr(ch,footerstring))
      {
        lc=-1;
        continue;
      }

      if(strstr(ch,totalstring))
      {
        ttemp[ii]='\r';
        ttemp[ii + 1]='\n';
        ttemp[ii + 2]='\0';
        fputs(ttemp,out);
        break;
      }

      if(ch[0] ==  '\r') continue;

      bc=0;      /* zani teti parametre od zaetka */
      pp=0;
      ss=0;
      while(ch[pp+1] != '\r')
      {

        pp++;
        if(ch[pp] == ' ' && ch[pp-1] == ' ') continue;
        else
        {
          if(!ss)
          {
            ss=1;      /* nova vrstica - prvi znak po presledkih je e bil */
            if(isdigit(ch[pp]))
            {
              if(ii)
              {
                ttemp[ii]='\r';
                ttemp[ii + 1]='\n';
                ttemp[ii + 2]='\0';
                fputs(ttemp,out);
              }
              ii=0;      /* zani pisati novo vrstico*/
            }
          }
          if(ch[pp] == ' ')
          {
            if(bc < 2 && ch[pp-1] != ' ') bc++;
            if(bc > 2) ttemp[bc]=' ';
            if(bc == 2) bc=ii;
            ttemp[ii]=';';
          }
          else ttemp[ii]=ch[pp];
          ii++;
        }
      }
    }

    fclose(in);
    fclose(out);
    goto EXIT;
  }


  /* test if PS or EPS file */

  if(!ps && !eps)
  {
    printf("Input file is not in PS or EPS format!\n");
    status=1;
    goto EXIT;
  }


  /* processing trim */

  if(trim)
  {
    sprintf(command,"\"%s\" -q -dNOPAUSE -dBATCH -sDEVICE=bbox %s > %s 2>&1",gsexec,infile,bbfile);
    if(!debug) printf("Calculating Trimmed Bounding Box ...\n");
    else printf("%s\n",command);
    status=system(command);
    if(status == -1) goto EXIT;

    if((in=fopen(bbfile,"r")) == NULL)
    {
      printf("Can't find %s!", bbfile);
      status=1;
      goto EXIT;
    }

    while ((fgets(ch,MAX_IN,in)) !=NULL)
    {
      if(debug) printf("%s",ch);
      if (strncmp(ch,bbstring,strlen(bbstring)) == 0 )
      {
        sscanf(ch+strlen(bbstring),"%lf %lf %lf %lf",&dbbox[0],&dbbox[1],&dbbox[2],&dbbox[3]);
        if(debug)printf("Original BoundingBox is: %.0f %.0f %.0f %.0f.\n",dbbox[0],dbbox[1],dbbox[2],dbbox[3]);
        if(border)
        {
          border=72*border;
          dbbox[0]=dbbox[0]-border;
          dbbox[1]=dbbox[1]-border;
          dbbox[2]=dbbox[2]+border;
          dbbox[3]=dbbox[3]+border;
        }
        for(i=0;i<4;i++) bbox[i]=dbbox[i];
        break;
      }
    }

    fclose(in);
    if(debug)printf("Trimed BoundingBox is: %.0f %.0f %.0f %.0f.\n",dbbox[0],dbbox[1],dbbox[2],dbbox[3]);
    else remove(bbfile);
  }


  /* process Windows PS print file */

  if(ps && !(pcb || sch))
  {
    in=fopen(infile, "r");
    out=fopen(tempfile,"w");

    while ((fgets(ch,MAX_IN,in)) !=NULL)
    {

      /* change bounding box */

      if (!strncmp(ch,bbstring,strlen(bbstring)))
      {
        if(!trim)   sscanf(ch+strlen(bbstring),"%lf %lf %lf %lf",&dbbox[0],&dbbox[1],&dbbox[2],&dbbox[3]);
          width   = dbbox[2]-dbbox[0];
          height  = dbbox[3]-dbbox[1];
          fprintf(out,"%s 0 0 %d %d\n", bbstring,width,height);
          BBCorrected=3;
          continue;
      }

      /* change PageSize */

      if(strstr(ch,pagesizestring))
      {
        fprintf(out,"    %s%d %d] put\n", pagesizestring,width,height);
        BBCorrected--;
        continue;
      }

      /* change coordinate system - /mysetup */

      if(!strncmp(ch,mystring,strlen(mystring)))
      {
        sscanf(ch+strlen(mystring),"%lf %lf %lf %lf %lf %lf", &temp[0],&temp[1],&temp[2],&temp[3],&temp[4],&temp[5]);
        temp[4]=temp[4]-dbbox[0];
        temp[5]=temp[5]-dbbox[1];
        fprintf(out,"%s %.3f %.3f %.3f %.3f %.3f %.3f ] |\n", mystring,temp[0],temp[1],temp[2],temp[3],temp[4],temp[5]);
        BBCorrected--;
        continue;
      }

      /* copy unchanged lines */

      fputs(ch,out);

    }
    fclose(in);
    fclose(out);
    if(BBCorrected != 1)BBCorrected=0;
  }


  /* process Tango file */

  page=0;
  if(pcb || sch)
  {

    if(sch)
    {
      newthick=thick*newthick;
      if(thin)newthin=thin*newthin;
      else newthin=thick*newthin;
      if(newthin > newthick) newthick=newthin;
      if(debug){
        printf("Thick is %.0f, thin is %.0f\n", thick,thin);
        printf("New thick is %0.1f, new thin is %0.1f\n", newthick,newthin);
      }
    }

    in=fopen(infile, "r");
    out=fopen(tempfile,"w");
    layer=0;
    layer_temp=layer_1st;

    while ((fgets(ch,MAX_IN,in)) !=NULL)
    {

      /* put layers of tango pcb file into wanted order */

      if(pcb)
      {
        if(ftell(in) == layer_temp)
        {
          fseek(in,laytable[layer].layer_start,SEEK_SET);
          layer_temp=laytable[layer].layer_end;
          layer++;
        }
      }


      /* change tango font */

      if(strstr(ch, oldfontstring) != NULL)
      {
        sstart=strstr(ch, oldfontstring)-ch;
        soffset=strlen(newfontstring)-strlen(oldfontstring);
        for(i=0;ch[i] != '\0' ;i++)
        {
          if(i < sstart)ttemp[i]=ch[i];
          else if(i+soffset >= sstart)ttemp[i+soffset]=ch[i];
        }
        for(i=0;newfontstring[i] != '\0';i++)ttemp[i+sstart]=newfontstring[i];
        fprintf(out,"%s", ttemp);
        continue;
      }


      /* change bounding box */

      if (strncmp(ch,bbstring,strlen(bbstring)) == 0 )
      {
        if(!trim)
        {
          sscanf(ch+strlen(bbstring),"%lf %lf %lf %lf",&dbbox[0],&dbbox[1],&dbbox[2],&dbbox[3]);
        }
        width   = dbbox[2]-dbbox[0];
        height  = dbbox[3]-dbbox[1];
        fprintf(out,"%s 0 0 %d %d\n", bbstring,width,height);
        fprintf(out,"<< /PageSize [%d %d] >> setpagedevice \n",width,height);
        continue;
      }


      /* change coordinate system */

      if(strncmp(ch,mulstring,strlen(mulstring)) == 0)
      {
        sscanf(ch+strlen(mulstring),"%lf",&mul);
      }
      if(strncmp(ch,pagestring,strlen(pagestring)) == 0) page=1;
      if(page && trim)
      {
        if(strstr(ch, "translate") != NULL){
          sscanf(ch,"%lf %s %lf",&dbbox[2],ttemp,&dbbox[3]);
          xoffset =dbbox[2]-dbbox[0]/mul;
          yoffset =dbbox[3]-dbbox[1]/mul;
          fprintf(out,"%d m %d m translate\n", xoffset, yoffset);
          continue;
        }
      }

      /* repairing square pad color - removing 1 graylevel from /sr definition */

      if(pcb)
      {
        if(strstr(ch, srstring) != NULL) sr=1;
        if(sr && strstr(ch, graystring))
        {
          sr=0;
          continue;
        }
      }

      /* making colors from gray levels */

      if(page)
      {
        if(tgocol)
        {
          if(strstr(ch,graystring) != NULL)
          {
            sscanf(ch+2,"%d", &i);
            i=i/5;
            rout=coltable[i].red;gout=coltable[i].green;bout=coltable[i].blue;
            fprintf(out,"%.2f %.2f %.2f %s\n", rout,gout,bout,colorstring);
            continue;
          }
        }
      }

      /* changing line width */

      if(sch)
      {
        if(strstr(ch, prstring) != NULL) pr=1;
        if(pr)
        {
          if(strstr(ch,"a5 3 eq"))
          {
            fprintf(out,"          {a5 3 eq {%.1f} {%.1f} ifelse "
                       "m setlinewidth stroke} ifelse\n", newthick, newthin);
            pr=0;
            continue;
          }
        }

        if(strstr(ch," pr") != NULL)
        {
          sscanf(ch,"%lf %lf %lf %lf", &temp[0],&temp[1],&temp[2],&temp[3]);
          fprintf(out,"%.0f %.0f %.0f %.0f 4 pr\n", temp[0],temp[1],temp[2],temp[3]);
          continue;
        }

        if(strstr(ch," pt") != NULL)
        {
          sscanf(ch,"%lf %lf %lf %lf %lf", &temp[0],&temp[1],&temp[2],&temp[3],&lwidth);
          if(lwidth == thick)lwidth = newthick;
          else lwidth = newthin;
          fprintf(out,"%.0f %.0f %.0f %.0f %.1f pt\n", temp[0],temp[1],temp[2],temp[3],lwidth);
          continue;
        }

        if(strstr(ch," pa") != NULL)
        {
          sscanf(ch,"%lf %lf %lf %lf %lf %lf", &temp[0],&temp[1],&lwidth,&temp[2],&temp[3],&temp[4]);
          if(lwidth == thick)lwidth = newthick;
          else lwidth = newthin;

          fprintf(out,"%.0f %.0f %.1f %.0f %.0f %.0f pa\n", temp[0],temp[1],lwidth,temp[2],temp[3],temp[4]);
          continue;
        }

      }

      /* copying unchanged line */

       fputs(ch,out);

    }
    fclose(in);
    fclose(out);
    BBCorrected=1;
  }


  if(!BBCorrected)
  {
    printf("Converting %s to %s ..... ",infile,outfile);
    if(debug) printf("\nAdding correct pagesize to EPS ...\n");
    in=fopen(infile, "r");
    out=fopen(tempfile,"w");
    while ((fgets(ch,MAX_IN,in)) !=NULL)
    {
      if (strncmp(ch,bbstring,strlen(bbstring)) == 0 )
      {
        BBCorrected=1;
        if(!trim)
        {
          sscanf(ch+strlen(bbstring),"%lf %lf %lf %lf",&dbbox[0],&dbbox[1],&dbbox[2],&dbbox[3]);
          for(i=0;i<4;i++) bbox[i]=dbbox[i];
        }
        if(debug) printf("%s %d %d %d %d\n", bbstring,bbox[0],bbox[1],bbox[2],bbox[3]);
        width   = bbox[2]-bbox[0];
        height  = bbox[3]-bbox[1];
        xoffset =-bbox[0];
        yoffset =-bbox[1];
        fprintf(out,"%s %d %d %d %d\n", bbstring,0,0,width,height);
        fprintf(out,"<< /PageSize [%d %d] >> setpagedevice \n",width,height);
        fprintf(out,"gsave %d %d translate\n",xoffset,yoffset);
        did_CR=1;
		  continue;
      }
      else
      {
        fputs(ch,out);
        if(strpbrk(ch, "\n")) did_CR=1;
        else                  did_CR=0;
      }
    }
    if(BBCorrected)
    {
      if(!did_CR) fprintf(out,"\ngrestore\n");
      else        fprintf(out,  "grestore\n");
    }
    fclose(in);
    fclose(out);
  }


  if((out=fopen(outfile,"w")) == NULL)
  {
    printf("Cannot open output file!\n");
    printf("You cannot simultaneously preview picture in Adobe Acrobat\n");
    printf("and generate new picture with the same name!\n");
    remove(tempfile);
    status=1;
    goto EXIT;
  }
  fclose(out);


  if(dopdf || dopng)
  {
    if(dopdf)
    {
      if(!compress) copt="-dUseFlateCompression=false ";
      else          copt="-dUseFlateCompression=true ";
      printf("%s \n", gsexec);
		sprintf(command,"%s -q -dNOPAUSE -dBATCH -dCompatibilityLevel=1.2 %s"
                      "-dColorImageResolution=300 -sDEVICE=pdfwrite -sOutputFile=%s  -c save pop -f %s",
                      gsexec,copt,outfile,tempfile);
    }
    else
    {
      if(dopng)
      {
        if(!pngxfactor)  pngxdimm=72*pngxdimm/width;
        sprintf(command,"%s -q -dNOPAUSE -dBATCH"
                        " -sDEVICE=%s -r%.0f -sOutputFile=%s %s",
                        gsexec,pngcol,pngxdimm,outfile,tempfile);
      }
    }
    if(debug) printf("%s\n", command);
    else      printf("Running ghostscript ...\n");
    status=system(command);
    if(debug)
    {
      if(trim)printf("Leaving temporary files: %s and %s.\n", bbfile,tempfile);
      else printf("Leaving temporary file: %s.\n", tempfile);
    }
    else
    {
      remove(tempfile);
      if(trim) remove(bbfile);
    }
  }
  else
  {
    rename(tempfile,outfile);
    if(debug && trim) printf("Leaving temporary file: %s.\n", bbfile);
  }
  printf("Done!\n");

EXIT:
  return status;
}

